@@ -1,13 +1,20 @@
 """Miscellaneous helpers live here"""
+import collections
+from contextlib import contextmanager
+import getpass
+import inspect
 import json
 import logging
 import os
+import pickle
 import sys
+import tarfile
 import time
 from collections import UserDict, namedtuple
 from collections.abc import MutableMapping
 from copy import deepcopy
 from pathlib import Path
+from uuid import uuid4
 
 import yaml
 from logzero import logger
@@ -76,7 +83,7 @@ def flatten_dict(nested_dict, parent_key="", separator="_"):
                 if isinstance(val, dict):
                     flattened.extend(flatten_dict(val, new_key, separator).items())
                     to_remove.append(index)
-            for index in to_remove:
+            for index in to_remove[::-1]:  # remove from back to front
                 del value[index]
             flattened.append((new_key, value))
         else:
@@ -173,25 +180,30 @@ def resolve_file_args(broker_args):
     then attempt to resolve them. If not resolved, keep arg/value pair intact.
     """
     final_args = {}
+    # parse the eventual args_file first
+    if val := broker_args.pop('args_file', None):
+        if isinstance(val, Path) or (
+            isinstance(val, str) and val[-4:] in ("json", "yaml", ".yml")
+        ):
+            if data := load_file(val):
+                if isinstance(data, dict):
+                    final_args.update(data)
+                elif isinstance(data, list):
+                    for d in data:
+                        final_args.update(d)
+            else:
+                raise exceptions.BrokerError(f"No data loaded from {val}")
+
     for key, val in broker_args.items():
         if isinstance(val, Path) or (
             isinstance(val, str) and val[-4:] in ("json", "yaml", ".yml")
         ):
             if data := load_file(val):
-                if key == "args_file":
-                    if isinstance(data, dict):
-                        final_args.update(data)
-                    elif isinstance(data, list):
-                        for d in data:
-                            final_args.update(d)
-                else:
-                    final_args[key] = data
-            elif key == "args_file":
-                raise exceptions.BrokerError(f"No data loaded from {val}")
+                final_args.update({key: data})
             else:
-                final_args[key] = val
+                final_arg.update({key: val})
         else:
-            final_args[key] = val
+            final_args.update({key: val})
     return final_args
 
 
@@ -322,16 +334,28 @@ class MockStub(UserDict):
         return self
 
     def __getitem__(self, key):
-        item = getattr(self, key, self)
+        if isinstance(key, str):
+            item = getattr(self, key, self)
         try:
             item = super().__getitem__(key)
         except KeyError:
-            pass
+            item = self
         return item
 
     def __call__(self, *args, **kwargs):
         return self
 
+    def __hash__(self):
+        return hash(
+            tuple(
+                (
+                    kp
+                    for kp in self.__dict__.items()
+                    if isinstance(kp[1], collections.abc.Hashable)
+                )
+            )
+        )
+
 
 def update_log_level(ctx, param, value):
     silent = False
@@ -446,3 +470,142 @@ class FileLock:
 
     def __exit__(self, *tb_info):
         self.return_file()
+
+
+class Result:
+    """Dummy result class for presenting results in dot access"""
+
+    def __init__(self, **kwargs):
+        self.__dict__.update(kwargs)
+
+    def __repr__(self):
+        return f"stdout:\n{self.stdout}\nstderr:\n{self.stderr}\nstatus: {self.status}"
+
+    @classmethod
+    def from_ssh(cls, stdout, channel):
+        return cls(
+            stdout=stdout,
+            status=channel.get_exit_status(),
+            stderr=channel.read_stderr(),
+        )
+
+    @classmethod
+    def from_duplexed_exec(cls, duplex_exec):
+        if duplex_exec.output[0]:
+            stdout = duplex_exec.output[0].decode("utf-8")
+        else:
+            stdout = ""
+        if duplex_exec.output[1]:
+            stderr = duplex_exec.output[1].decode("utf-8")
+        else:
+            stderr = ""
+        return cls(
+            status=duplex_exec.exit_code,
+            stdout=stdout,
+            stderr=stderr,
+        )
+
+    @classmethod
+    def from_nonduplexed_exec(cls, nonduplex_exec):
+        return cls(
+            status=nonduplex_exec.exit_code,
+            stdout=nonduplex_exec.output.decode("utf-8"),
+            stderr="",
+        )
+
+
+def find_origin():
+    """Move up the call stack to find tests, fixtures, or cli invocations.
+    Additionally, return the jenkins url, if it exists.
+    """
+    prev, jenkins_url = None, os.environ.get("BUILD_URL")
+    for frame in inspect.stack():
+        if frame.function == "checkout" and frame.filename.endswith(
+            "broker/commands.py"
+        ):
+            return f"broker_cli:{getpass.getuser()}", jenkins_url
+        if frame.function.startswith("test_"):
+            return f"{frame.function}:{frame.filename}", jenkins_url
+        if frame.function == "call_fixture_func":
+            # attempt to find the test name from the fixture's request object
+            if request := _frame.frame.f_locals.get("request"):
+                return f"{prev} for {request.node._nodeid}", jenkins_url
+            # otherwise, return the fixture name and filename
+            return prev or "Uknown fixture", jenkins_url
+        prev, _frame = f"{frame.function}:{frame.filename}", frame
+    return f"Unknown origin by {getpass.getuser()}", jenkins_url
+
+
+@contextmanager
+def data_to_tempfile(data, path=None, as_tar=False):
+    """Write data to a temporary file and return the path"""
+    path = Path(path or uuid4().hex[-10])
+    logger.debug(f"Creating temporary file {path.absolute()}")
+    if isinstance(data, bytes):
+        path.write_bytes(data)
+    elif isinstance(data, str):
+        path.write_text(data)
+    else:
+        raise TypeError(f"data must be bytes or str, not {type(data)}")
+    if as_tar:
+        tar = tarfile.open(path)
+        yield tarfile.open(path)
+        tar.close()
+    else:
+        yield path
+    path.unlink()
+
+
+@contextmanager
+def temporary_tar(paths):
+    """Create a temporary tar file and return the path"""
+    temp_tar = Path(f"{uuid4().hex[-10]}.tar")
+    with tarfile.open(temp_tar, mode="w") as tar:
+        for path in paths:
+            logger.debug(f"Adding {path.absolute()} to {temp_tar.absolute()}")
+            tar.add(path, arcname=path.name)
+    yield temp_tar.absolute()
+    temp_tar.unlink()
+
+
+class PickleSafe:
+    """A class that helps with pickling and unpickling complex objects"""
+
+    def _pre_pickle(self):
+        """This method is called before pickling an object"""
+        pass
+
+    def _post_pickle(self, purified):
+        """This method is called after pickling an object
+
+        purified will be a list of names of attributes that were removed
+        """
+        pass
+
+    def _purify(self):
+        """Strip all unpickleable attributes from a Host before pickling"""
+        self.purified = getattr(self, "purified", [])
+        for name in list(self.__dict__):
+            self._purify_target = name
+            try:
+                pickle.dumps(self.__dict__[name])
+            except (pickle.PicklingError, AttributeError):
+                self.__dict__[name] = None
+                self.purified.append(name)
+
+    def __getstate__(self):
+        """If a session is active, remove it for pickle compatability"""
+        self._pre_pickle()
+        try:
+            self._purify()
+        except RecursionError:
+                logger.warning(f"Recursion limit reached on {self._purify_target}")
+                self.__dict__[self._purify_target] = None
+                self.__getstate__()
+        self.__dict__.pop("_purify_target", None)
+        return self.__dict__
+
+    def __setstate__(self, state):
+        """Sometimes pickle strips things that we need. This should be used to restore them"""
+        self.__dict__.update(state)
+        self._post_pickle(purified=getattr(self, "purified", []))
