@@ -1,10 +1,13 @@
+import pickle
 import dynaconf
 
 from broker import exceptions
+from broker.helpers import PickleSafe
 from broker.settings import settings
+from logzero import logger
 
 
-class Provider:
+class Provider(PickleSafe):
     # Populate with a list of Dynaconf Validators specific to your provider
     _validators = []
     # Set to true if you don't want your provider shown in the CLI
@@ -14,8 +17,12 @@ class Provider:
     _checkout_options = []
     _execute_options = []
 
-    def __init__(self):
+    def __init__(self, **kwargs):
         self._construct_params = []
+        cls_name = self.__class__.__name__
+        logger.debug(f"{cls_name} provider instantiated with {kwargs=}")
+        instance_name = kwargs.pop(f"{cls_name}", None)
+        self._validate_settings(instance_name)
 
     def _validate_settings(self, instance_name=None):
         """Load and validate provider settings
@@ -27,46 +34,36 @@ class Provider:
 
         :param instance_name: A string matching an instance name
         """
+        instance_name = instance_name or getattr(self, "instance", None)
         section_name = self.__class__.__name__.upper()
         # make sure each instance isn't loading values from another
         fresh_settings = settings.get_fresh(section_name)
         instance, default = None, False
         for candidate in fresh_settings.instances:
+            logger.debug(f"Checking {instance_name} against {candidate}")
             if instance_name in candidate:
                 instance = candidate
                 default = False
+                break
             elif (
                 candidate.values()[0].get("default")
                 or len(fresh_settings.instances) == 1
             ):
                 instance = candidate
                 default = True
+        self.instance, *_ = instance  # store the instance name on the provider
         fresh_settings.update(instance.values()[0])
+        settings[section_name] = fresh_settings
         if default:
             # if a default provider is selected, defer to loaded environment variables
-            # settings[section_name] = fresh_settings
-            # settings.execute_loaders(loaders=[dynaconf.loaders.env_loader])
-            # ideal solution above. However, need to workaround until
-            # https://github.com/rochacbruno/dynaconf/issues/511
-            settings.execute_loaders()
-            for key in fresh_settings.keys():
-                if key in settings[section_name]:
-                    fresh_settings[key] = settings[section_name][key]
-            settings[section_name] = fresh_settings
-        else:
-            settings[section_name] = fresh_settings
+            settings.execute_loaders(loaders=[dynaconf.loaders.env_loader])
 
-        # temporary workaround for https://github.com/rochacbruno/dynaconf/issues/508
-        # remove the current valiators, add ours, and validate
-        # then add the other validators back in and move on
-        current_validators = settings.validators[:]
-        settings.validators.clear()
         settings.validators.extend(self._validators)
+        # use selective validation to only validate the instance settings
         try:
-            settings.validators.validate()
+            settings.validators.validate(only=section_name)
         except dynaconf.ValidationError as err:
             raise exceptions.ConfigurationError(err)
-        settings.validators.extend(current_validators)
 
     def _host_release(self):
         raise exceptions.NotImplementedError("_host_release has not been implemented")
@@ -95,3 +92,26 @@ class Provider:
 
     def release(self, host_obj):
         raise exceptions.NotImplementedError("release has not been implemented")
+
+    def __repr__(self):
+        inner = ", ".join(
+            f"{k}={v}"
+            for k, v in self.__dict__.items()
+            if not k.startswith("_") and not callable(v)
+        )
+        return f"{self.__class__.__name__}({inner})"
+
+    def __getstate__(self):
+        """If a session is active, remove it for pickle compatability"""
+        self._purify()
+        return self.__dict__
+
+    def _purify(self):
+        """Strip all unpickleable attributes from a Host before pickling"""
+        for key, obj in self.__dict__.items():
+            try:
+                pickle.dumps(obj)
+            except (pickle.PicklingError, AttributeError):
+                self.__dict__[key] = None
+            except RecursionError:
+                logger.warning(f"Recursion limit reached on {obj=}")
