@@ -3,10 +3,10 @@ import inspect
 import json
 import yaml
 from urllib import parse as url_parser
-from functools import cached_property
+from functools import cache, cached_property
 from dynaconf import Validator
 from broker import exceptions
-from broker.helpers import results_filter
+from broker.helpers import find_origin, results_filter
 from broker.settings import settings
 from logzero import logger
 from datetime import datetime
@@ -22,6 +22,51 @@ from broker.providers import Provider
 from broker import helpers
 
 
+@cache
+def get_awxkit_and_uname(
+    config=None, root=None, url=None, token=None, uname=None, pword=None
+):
+    """Return an awxkit api object and resolved username"""
+    # Prefer token if its set, otherwise use username/password
+    # auth paths for the API taken from:
+    # https://github.com/ansible/awx/blob/ddb6c5d0cce60779be279b702a15a2fddfcd0724/awxkit/awxkit/cli/client.py#L85-L94
+    # unit test mock structure means the root API instance can't be loaded on the same line
+    config = config or awxkit.config
+    config.base_url = url
+    if root is None:
+        root = awxkit.api.Api()  # support mock stub for unit tests
+    if token:
+        helpers.emit(auth_type="token")
+        logger.info("Using token authentication")
+        config.token = token
+        try:
+            root.connection.login(
+                username=None, password=None, token=token, auth_type="Bearer"
+            )
+        except awxkit.exceptions.Unauthorized as err:
+            raise exceptions.AuthenticationError(err.args[0])
+        versions = root.get().available_versions
+        try:
+            # lookup the user that authenticated with the token
+            # If a username was specified in config, use that instead
+            my_username = uname or versions.v2.get().me.get().results[0].username
+        except (IndexError, AttributeError):
+            # lookup failed for whatever reason
+            raise exceptions.ProviderError(
+                provider="AnsibleTower",
+                message="Failed to lookup a username for the given token, please check credentials",
+            )
+    else:  # dynaconf validators should have checked that either token or password was provided
+        helpers.emit(auth_type="password")
+        logger.info("Using username and password authentication")
+        config.credentials = {"default": {"username": uname, "password": pword}}
+        config.use_sessions = True
+        root.load_session().get()
+        versions = root.available_versions
+        my_username = uname
+    return versions.v2.get(), my_username
+
+
 class AnsibleTower(Provider):
 
     _validators = [
@@ -73,11 +118,7 @@ class AnsibleTower(Provider):
     ]
 
     def __init__(self, **kwargs):
-        # Get the specified AT instance
-        logger.debug(f"AnsibleTower instantiated with {kwargs=}")
-        instance_name = kwargs.pop("AnsibleTower", None)
-        # Validate the AnsibleTower-specific settings
-        self._validate_settings(instance_name)
+        super().__init__(**kwargs)
         # get our instance settings
         self.url = settings.ANSIBLETOWER.base_url
         self.uname = settings.ANSIBLETOWER.get("username")
@@ -87,51 +128,18 @@ class AnsibleTower(Provider):
             kwargs.get("tower_inventory") or settings.ANSIBLETOWER.inventory
         )
         # Init the class itself
-        self._construct_params = []
-        config = kwargs.get("config", awxkit.config)
-        config.base_url = self.url
+        config = kwargs.get("config")
         root = kwargs.get("root")
-        if root is None:
-            root = awxkit.api.Api()  # support mock stub for unit tests
-        # Prefer token if its set, otherwise use username/password
-        # auth paths for the API taken from:
-        # https://github.com/ansible/awx/blob/ddb6c5d0cce60779be279b702a15a2fddfcd0724/awxkit/awxkit/cli/client.py#L85-L94
-        # unit test mock structure means the root API instance can't be loaded on the same line
-        if self.token:
-            helpers.emit(auth_type="token")
-            logger.info("Using token authentication")
-            config.token = self.token
-            try:
-                root.connection.login(
-                    username=None, password=None, token=self.token, auth_type="Bearer"
-                )
-            except awxkit.exceptions.Unauthorized as err:
-                raise exceptions.AuthenticationError(err.args[0])
-            versions = root.get().available_versions
-            try:
-                # lookup the user that authenticated with the token
-                # If a username was specified in config, use that instead
-                my_username = (
-                    self.uname or versions.v2.get().me.get().results[0].username
-                )
-            except (IndexError, AttributeError):
-                # lookup failed for whatever reason
-                raise exceptions.ProviderError(
-                    provider="AnsibleTower",
-                    message="Failed to lookup a username for the given token, please check credentials",
-                )
-        else:  # dynaconf validators should have checked that either token or password was provided
-            helpers.emit(auth_type="password")
-            logger.info("Using username and password authentication")
-            config.credentials = {
-                "default": {"username": self.uname, "password": self.pword}
-            }
-            config.use_sessions = True
-            root.load_session().get()
-            versions = root.available_versions
-            my_username = self.uname
-        self.v2 = versions.v2.get()
-        self.username = my_username
+        self.v2, self.username = get_awxkit_and_uname(
+            config=config,
+            root=root,
+            url=self.url,
+            token=self.token,
+            uname=self.uname,
+            pword=self.pword,
+        )
+        # Check to see if we're running AAP (ver 4.0+)
+        self._is_aap = False if self.v2.ping.get().version[0] == "3" else True
 
     @staticmethod
     def _pull_params(kwargs):
@@ -152,6 +160,14 @@ class AnsibleTower(Provider):
         broker_args = getattr(caller_host, "_broker_args", {}).get("_broker_args", {})
         # remove the workflow field since it will conflict with the release workflow
         broker_args.pop("workflow", None)
+        # reconstruct tower provider inventory information when needed
+        prov_inv = broker_args.pop("tower_inventory", None)
+        if isinstance(prov_inv, str):
+            logger.debug(f"prov_inv: {prov_inv}")
+            prov_inv = self._translate_inventory(prov_inv)
+        if prov_inv:
+            logger.debug(f"prov_inv: {prov_inv}")
+            broker_args["inventory"] = prov_inv
         caller_host._prov_inst.release(
             broker_args.get("source_vm", caller_host.name), broker_args
         )
@@ -166,6 +182,30 @@ class AnsibleTower(Provider):
             }
         )
 
+    def _translate_inventory(self, inventory):
+        if isinstance(inventory, int):  # already an id, silly
+            if inventory_info := self.v2.inventory.get(id=inventory):
+                return inventory_info.results[0].name
+            else:
+                raise exceptions.ProviderError(
+                    provider="AnsibleTower",
+                    message=f"Unknown AnsibleTower inventory by id {inventory}",
+                )
+        if inventory_info := self.v2.inventory.get(search=inventory):
+            if inventory_info.count > 1:
+                raise exceptions.ProviderError(
+                    provider="AnsibleTower",
+                    message=f"Ambigious AnsibleTower inventory name {inventory}",
+                )
+            elif inventory_info.count == 1:
+                inv_struct = inventory_info.results.pop()
+                return inv_struct.id
+            else:
+                raise exceptions.ProviderError(
+                    provider="AnsibleTower",
+                    message=f"Unknown AnsibleTower inventory {inventory}",
+                )
+
     def _merge_artifacts(self, at_object, strategy="last", artifacts=None):
         """Gather and merge all artifacts associated with an object and its children
 
@@ -300,7 +340,9 @@ class AnsibleTower(Provider):
             "name": host.name,
             "type": host.type,
             "hostname": host.variables.get("fqdn"),
+            "tower_inventory": self._translate_inventory(host.inventory),
             "_broker_provider": "AnsibleTower",
+            "_broker_provider_instance": self.instance,
             "_broker_args": getattr(host, "_broker_args", {}),
         }
         expire_time = self._get_expire_date(host.id)
@@ -334,7 +376,6 @@ class AnsibleTower(Provider):
                 if val and isinstance(val, str)
             }
         )
-        host_info["_broker_args"]["tower_inventory"] = create_job.inventory
         return host_info
 
     @cached_property
@@ -344,20 +385,8 @@ class AnsibleTower(Provider):
         elif isinstance(self._inventory, int):
             # inventory already resolved as id
             return self._inventory
-        if inventory_info := self.v2.inventory.get(search=self._inventory):
-            if inventory_info.count > 1:
-                raise exceptions.ProviderError(
-                    provider="AnsibleTower",
-                    message=f"Ambigious AnsibleTower inventory name {self._inventory}",
-                )
-            elif inventory_info.count == 1:
-                inv_struct = inventory_info.results.pop()
-                return inv_struct.id
-            else:
-                raise exceptions.ProviderError(
-                    provider="AnsibleTower",
-                    message=f"Unknown AnsibleTower inventory {self._inventory}",
-                )
+        self._inventory = self._translate_inventory(self._inventory)
+        return self._inventory
 
     def construct_host(self, provider_params, host_classes, **kwargs):
         """Constructs host to be read by Ansible Tower
@@ -394,7 +423,7 @@ class AnsibleTower(Provider):
                 if key.endswith("host_type"):
                     host_type = value if value in host_classes else host_type
             if not hostname:
-                raise Exception(f"No hostname found in job attributes:\n{job_attrs}")
+                logger.warning(f"No hostname found in job attributes:\n{job_attrs}")
             logger.debug(f"hostname: {hostname}, name: {name}, host type: {host_type}")
             host_inst = host_classes[host_type](
                 **{**kwargs, "hostname": hostname, "name": name}
@@ -414,6 +443,10 @@ class AnsibleTower(Provider):
         if name := kwargs.get("workflow"):
             subject = "workflow"
             get_path = self.v2.workflow_job_templates
+            origin = find_origin()
+            kwargs["_broker_origin"] = origin[0]
+            if origin[1]:
+                kwargs["_jenkins_url"] = origin[1]
         elif name := kwargs.get("job_template"):
             subject = "job_template"
             get_path = self.v2.job_templates
@@ -432,11 +465,14 @@ class AnsibleTower(Provider):
                 provider="AnsibleTower",
                 message=f"{subject.capitalize()} not found by name: {name}",
             )
-        payload = {"extra_vars": str(kwargs)}
-        if self.inventory:
+        payload = {}
+        if inventory := kwargs.pop("inventory", None):
+            payload["inventory"] = inventory
+        elif self.inventory:
             payload["inventory"] = self.inventory
         else:
             logger.info("No inventory specified, Ansible Tower will use a default.")
+        payload["extra_vars"] = str(kwargs)
         logger.debug(
             f"Launching {subject}: {url_parser.urljoin(self.url, str(target.url))}\n"
             f"{payload=}"
@@ -444,7 +480,12 @@ class AnsibleTower(Provider):
         job = target.launch(payload=payload)
         job_number = job.url.rstrip("/").split("/")[-1]
         job_api_url = url_parser.urljoin(self.url, str(job.url))
-        job_ui_url = url_parser.urljoin(self.url, f"/#/{subject}s/{job_number}")
+        if self._is_aap:
+            job_ui_url = url_parser.urljoin(
+                self.url, f"/#/jobs/{subject}/{job_number}/output"
+            )
+        else:
+            job_ui_url = url_parser.urljoin(self.url, f"/#/{subject}s/{job_number}")
         helpers.emit(api_url=job_api_url, ui_url=job_ui_url)
         logger.info("Waiting for job: \n" f"API: {job_api_url}\n" f"UI: {job_ui_url}")
         job.wait_until_completed(timeout=settings.ANSIBLETOWER.workflow_timeout)
@@ -512,6 +553,7 @@ class AnsibleTower(Provider):
             ]
             if res_filter := kwargs.get("results_filter"):
                 workflows = results_filter(workflows, res_filter)
+                workflows = workflows if isinstance(workflows, list) else [workflows]
             workflows = "\n".join(workflows[:results_limit])
             logger.info(f"Available workflows:\n{workflows}")
         elif inventory := kwargs.get("inventory"):
@@ -525,6 +567,7 @@ class AnsibleTower(Provider):
             ]
             if res_filter := kwargs.get("results_filter"):
                 inv = results_filter(inv, res_filter)
+                inv = inv if isinstance(inv, list) else [inv]
             inv = "\n".join(inv[:results_limit])
             logger.info(f"Available Inventories:\n{inv}")
         elif job_template := kwargs.get("job_template"):
@@ -540,6 +583,7 @@ class AnsibleTower(Provider):
             ]
             if res_filter := kwargs.get("results_filter"):
                 job_templates = results_filter(job_templates, res_filter)
+                job_templates = job_templates if isinstance(job_templates, list) else [job_templates]
             job_templates = "\n".join(job_templates[:results_limit])
             logger.info(f"Available job templates:\n{job_templates}")
         elif kwargs.get("templates"):
@@ -554,6 +598,7 @@ class AnsibleTower(Provider):
             templates.sort(reverse=True)
             if res_filter := kwargs.get("results_filter"):
                 templates = results_filter(templates, res_filter)
+                templates = templates if isinstance(templates, list) else [templates]
             templates = "\n".join(templates[:results_limit])
             logger.info(f"Available templates:\n{templates}")
         else:
@@ -568,6 +613,14 @@ class AnsibleTower(Provider):
             **broker_args,
         )
 
+    def __repr__(self):
+        inner = ", ".join(
+            f"{k}={v}"
+            for k, v in self.__dict__.items()
+            if not k.startswith("_") and not callable(v)
+        )
+        return f"{self.__class__.__name__}({inner})"
+
 
 def awxkit_representer(dumper, data):
     """In order to resolve awxkit objects, a custom representer is needed"""
