@@ -19,12 +19,22 @@ validators = [
     Validator("HOST_USERNAME", default="root"),
     Validator("HOST_PASSWORD", must_exist=True),
     Validator("HOST_CONNECTION_TIMEOUT", default=None),
+    Validator("HOST_SSH_PORT", default=22),
+    Validator("HOST_SSH_KEY_FILENAME", default=None),
 ]
+
+# temportary fix for dynaconf #751
+vault_vars = {k: v for k, v in os.environ.items() if "VAULT_" in k}
+for k in vault_vars:
+    del os.environ[k]
+
 settings = Dynaconf(
     settings_file=str(settings_path.absolute()),
     ENVVAR_PREFIX_FOR_DYNACONF="BROKER",
     validators=validators,
 )
+# to make doubly sure, remove the vault loader if set somehow
+settings._loaders = [loader for loader in settings._loaders if not "vault" in loader]
 
 try:
     settings.validators.validate()
@@ -32,3 +42,5 @@ except ValidationError as err:
     raise ConfigurationError(
         f"Configuration error in {settings_path.absolute()}: {err.args[0]}"
     )
+
+os.environ.update(vault_vars)
